"""Extremely simple time-out queues for channels

These do *not* attempt to store/manage state, they *only*
queue up updates for clients who are temporarily unable to
retrieve them (e.g. because of network failures or regular
timeout/restart of connections).
"""
import bisect,weakref,time,logging
log = logging.getLogger( 'channels')

class _Highest( object ):
    """Utility object that sorts highest"""
    def __lt__( self, other ):
        return False
    def __gt__( self, other ):
        return True
HIGHEST = _Highest()

class Channel( object ):
    """A particular (active) channel being listened to by a user

    Messages should only be added to the channel in their final form,
    i.e. after all updates to the db...
    """
    all_channels = []
    all_channel_map = weakref.WeakValueDictionary()
    _clean_count = 0
    QUEUE_CLEAN_CYCLE = 2000
    _last_clean = 0.0
    _state = None

    @classmethod
    def get_existing( cls, key ):
        """Retrieve any existing channel for the given key"""
        return cls.all_channel_map.get( key )
    @classmethod
    def get_channel( cls, key, state=None ):
        """Find channel of the given key"""
        if cls._clean_count > cls.QUEUE_CLEAN_CYCLE:
            cls.clean()
            cls._clean_count = 0
        else:
            cls._clean_count += 1

        channel = cls.all_channel_map.get( key )
        if channel is not None:
            return channel
        channel = cls( key, state )
        cls.all_channels.append( channel )
        cls.all_channel_map[key] = channel
        return channel
    @classmethod
    def clean( cls, timeout=120 ):
        """Clean out all old messages and channels

        returns number of channels removed...
        """
        # Again, *not* thread-safe...
        oldts = time.time() - timeout
        new_channels = []
        for channel in cls.all_channels:
            channel._clean( oldts )
            if not channel._stale( oldts ):
                new_channels.append( channel )
        cls.all_channels[:] = new_channels
        # TODO: increase threhold if we haven't done anything this pass

    def __init__( self, key, state=None ):
        """Initialize this channel with the given key"""
        self.key = key
        self.queue = []
        self.watchers = []
        ts = self.ts()
        if state:
            self.set_state( state )

    def set_state( self, state, ts=None ):
        """Set our state value for "initial" values

        State should be a series of messages to the client
        which give them the current state of the channel.
        """
        if not ts:
            ts = self.ts()
        self._state = dict(
            ts=ts,
            messages=state or [],
            initial=True,
        )
        return self._state
    def get_state( self ):
        """Get our current state and it's ts value"""
        return self._state

    def add_watcher( self, watcher ):
        """Add a single watcher to the set of watchers"""
        self.watchers.append( watcher )

    def ts( self, record=True ):
        """Produce (and record) a timestamp"""
        t = time.time()
        if record:
            self._ts = t
        return t
    def send( self, message ):
        """Send message to this channel"""
#        info( 'Channel %s received %s', self.key, message )
        ts = self.ts()
        self.queue.append( (ts,message) )
        message['channel'] = self.key
        message['ts'] = ts

        new_watchers = []
        while self.watchers:
            watcher = self.watchers.pop()
            watcher( message )
            if not watcher.invalid:
                new_watchers.append( watcher )
        self.watchers.extend( new_watchers )

        return ts

    def since( self, ts ):
        """Return all messages since given ts

        This is to be called when something wants to check for updates,
        we'll have up to timeout seconds worth of previous messages and
        can (reasonably) efficiently check for any new messages...
        """
        # update our activity ts
        self.ts()
        # TODO: if we want to allow multi-threading, this isn't safe...
        # TODO: store the n-1'th timestamp so that if the user asks for
        # messages from before our queue we can report an error rather than
        # returning a partial response
        idx = bisect.bisect_left( self.queue, (ts,))
        return [m for ts,m in self.queue[idx:]]

    def _clean( self, oldts ):
        """Clean all messages older than timeout"""
        if self.queue:
            oldlen = len(self.queue)
            del self.queue[:bisect.bisect_right( self.queue,(oldts,HIGHEST))]
            newlen = len(self.queue)
            return oldlen - newlen
        return 0

    def _stale( self, oldts ):
        """Are we older than oldts?"""
        if self._ts < oldts:
            # no-one has asked for us in timeout seconds...
            return True
        else:
            return False
